-- PATH Class --

_ = require "underscore"

Pathfinder = {}
Pathfinder.__index = Pathfinder

setmetatable(Pathfinder, {
	__call = function (cls, ...)
		return cls.new(...)
		end,
	})

function Pathfinder.new(grid)
	local self = setmetatable({}, Pathfinder)

	self.moveready = false
	self.pathready = false

	self.path = {}
	self.nodes = grid


	-- set up grid node information
	for x = 1,#self.nodes do
		for y = 1, #self.nodes do
			local tile = self.nodes[x][y]
			self.nodes[x][y] = {}
			self.nodes[x][y].tile = tile

			if self.nodes[x][y].tile == 2 or self.nodes[x][y].tile == 3 then
				self.nodes[x][y].walkable = false
			else
				self.nodes[x][y].walkable = true
			end

			self.nodes[x][y].cost = 10
			self.nodes[x][y].F = 0
			self.nodes[x][y].x = x
			self.nodes[x][y].y = y
			self.nodes[x][y].parent = {}
		end
	end
	return self
end

function Pathfinder:find(cx,cy,ex,ey)

	if ex < 1 or ey < 1 or ex > #self.nodes or ey > #self.nodes then return nil end

	-- init and add starting node to the open list
	self.nodes[cx][cy].cost = 0
	self.nodes[cx][cy].F = manhattan(cx, cy, ex, ey)
	local open = {self.nodes[cx][cy]}
	-- init closed list
	local closed = {}

	local current = {}

	while not _.is_empty(open) do
		--sort by lowest F cost
		_.sort(open, function(a,b) return a.F < b.F end)
		-- take first off open list
		current = _.shift(open)
		-- switch current to closed list
		table.insert(closed, current)
		--if end square is in the closed table we're done
		if(table.contains(closed,self.nodes[ex][ey])) then
			break
		end

		neighbors = {}
		for x=-1, 1 do 
			for y=-1, 1 do
				local nx = current.x + x
				local ny = current.y + y
				if self.nodes[nx][ny] ~= nil then
					if (math.abs(x) == math.abs(y)) then --diagonal
						self.nodes[nx][ny].cost = 14
					else
						self.nodes[nx][ny].cost = 10
					end
					table.insert(neighbors, self.nodes[nx][ny])
				end
			end
		end

		-- For each adjacent square do
		for k,neighbor in pairs(neighbors) do
			-- if it's walkable and not in the closed list
			if neighbor.walkable and not table.contains(closed, neighbor) then
				
				local G = current.cost + neighbor.cost
				if not table.contains(open, neighbor) or G < current.cost then
					-- make the current square the parent of this square
					neighbor.parent = current

					H = manhattan(neighbor.x, neighbor.y, ex, ey)

					neighbor.cost = G 
					neighbor.F = neighbor.cost + H
					 	
					-- add it to the open list
					if not table.contains(open, neighbor) then
						table.insert(open, neighbor)	
					end
				 end
			end
		end

	end

	local path = {}

	if self.nodes[ex][ey].parent == nil then
		pathready = false
		self.path = nil
	else
		target = self.nodes[ex][ey]
		while target ~= self.nodes[cx][cy] do
			table.insert(path, {x=target.x,y=target.y})
			target = target.parent
		end
		table.insert(path, {x=cx,y=cy})
		pathready = true
		self.path = path
	end
end

function manhattan(currentx, currenty, targetx, targety)
	H = (10*math.abs(currentx-targetx) + 10*math.abs(currenty-targety))
	return H
end

function table.contains(table, element)
  for k, value in pairs(table) do
    if value == element then
      return true
    end
  end
  return false
end